#!/usr/bin/perl -w
#
# file    : index.pl
# date    : 2005-01-16
# author  : Bas Couwenberg <sebastic@xs4all.nl>
#

use strict qw(subs vars);
use Cwd;
use File::Basename ();
use Date::Calc qw(:all);
use Apache::Request;

my $r = Apache->request;	
my $c = $r->connection;
my $s = $r->server;

# Initialize global variables
my %args = &get_args($r, $c, $s);
   $args{"fci-version"} 	 	= '0.1.2';
   $args{"stanford-data-dir"}	 	= 'stanford-data/';
   $args{"client-data-dir"}	 	= 'client-data/';
  
   $args{"client-list-file"}	 	= $args{"client-data-dir"}.'client-list';
   $args{"project-list-file"}	 	= $args{"client-data-dir"}.'project-list';
   $args{"username-list-file"}	 	= $args{"client-data-dir"}.'username-list';

   $args{"user-summary-file"}	 	= $args{"stanford-data-dir"}.'daily_user_summary.txt';
   $args{"team-summary-file"}	 	= $args{"stanford-data-dir"}.'daily_team_summary.txt';
   $args{"project-summary-file"} 	= $args{"stanford-data-dir"}.'psummary.html';
   
   $args{"user-summary-update-file"}	= $args{"user-summary-file"}.'.lastupdate';
   $args{"team-summary-update-file"}	= $args{"team-summary-file"}.'.lastupdate';
   $args{"project-summary-update-file"} = $args{"project-summary-file"}.'.lastupdate';
   
   $args{action} 			= '' 		if(!$args{action});
   $args{target} 			= 'home.plc' 	if(!$args{target});


$r->header_out(Pragma => 'no-cache');
$r->header_out('Cache-control' => 'no-cache');
$r->header_out(Connection => 'close');

$r->content_type('text/html');
$r->send_http_header;

# If the UserAgent only requested the HTTP header (using HEAD perhaps), 
# we should send it and then exit, skipping the rest of the HTML document.

return 'OK' if $r->header_only;

my $sitename = 'Folding@Home Client Info';
	
htmlheader( title		 => $sitename,
	    link1       	 => 'stylesheet.css',
	    link1_rel   	 => 'stylesheet',
	    doctype              => 'HTML',
	    doctype_dtd          => 'HTML 4.01',
	    doctype_transitional => 1
	);


print "<!-- start main -->\n";

# check existance/permissions of direcotries used by 

my ($error, $msg) = (0, '');

my ($e1, $m1) = check_directory("$args{'client-data-dir'}", ('r', 'w'));
$error = 1   if($e1);
$msg  .= $m1 if($m1);

my ($e2, $m2) = check_directory("$args{'stanford-data-dir'}", ('r'));
$error = 1   if($e2);
$msg  .= $m2 if($m2);

if($error) {
	print "<pre>\nPermission Check Failed, see INSTALL file for required permissions\n\nErrors:\n$msg</pre>\n";
	htmlfooter();
	exit;
}

if($args{action} eq 'upload') {
	print "<pre>\n";
	if($args{unitinfo_file} ne '' && $args{config_file} && $args{client_name} ne '') {
		# save client name in list file of all known client names
		my $updated = update_client_list($args{'client-list-file'}, $args{client_name});
		print "Successfully added client to client-list\n" if($updated); 

		
		# Write general info file for this client
		my $info_file = $args{"client-data-dir"}.'generalinfo.'.$args{client_name};
		
		if(!-e $info_file || -w $info_file) {
			open(F, ">$info_file") || die "Cannot open file ($!)";
			print F "last_update:$args{Date} $args{Time}\n";
			print F "client_ip:$args{RemoteIP}\n";
			close F;
			
			print "Successfully saved general info file for client ($args{client_name})\n";
		}
		else {
			print "Error: saving general info file for client ($args{client_name}): Cannot write file!\n";
		}

		
		# Save unitinfo.txt file for the client
		my $unitinfo_file = $args{"client-data-dir"}.'unitinfo.'.$args{client_name};
		
		if(!-e $unitinfo_file || -w $unitinfo_file) {
			open(F, ">$unitinfo_file") || die "Cannot open file ($!)";
			print F $args{unitinfo_file};
			close F;

			print "Successfully saved unitinfo file for client ($args{client_name})\n";
		}
		else {
			print "Error: saving unitinfo file for client ($args{client_name}): Cannot write file!\n";
		}

		
		# Save client.cfg file for this client
		my $config_file = $args{"client-data-dir"}.'client-cfg.'.$args{client_name};
		
		if(!-e $config_file || -w $config_file) {
			open(F, ">$config_file") || die "Cannot open file ($!)";
			print F $args{config_file};
			close F;
			
			print "Successfully saved config file for client ($args{client_name})\n";
		}
		else {
			print "Error: saving config file for client ($args{client_name}): Cannot write file!\n";
		}


		# save project name in list file with all currently active projects
		   $updated  = update_project_list($args{'project-list-file'});
		print "Successfully added project to project-list\n" if($updated);

		
		# save username & client name in list file
		   $updated  = update_username_list($args{'username-list-file'});
		print "Successfully added username to username-list\n" if($updated);
	}
	else {
		if(!$args{unitinfo_file} || $args{unitinfo_file} eq '') {
			print "ERROR: No unitinfo file uploaded!\n";
		}
		elsif(!$args{config_file} || $args{config_file} eq '') {
			print "ERROR: No config file uploaded!\n";
		}
		elsif(!$args{client_name} || $args{client_name} eq '') {
			print "ERROR: No client name specified!\n";
		}
		else {
			print "ERROR: Fatal Exception!\n";
		}
	}
	print "</pre>\n";
}
else {
	print "<table align='center' width='95%' cellspacing='0' class='table-border'>\n";
	print "<tr><th colspan='2' class='table-head'>$sitename</th></tr>\n";

	print "<tr><td valign='top' class='right-border' width='20%'>\n\n";

	call('menu.plc', %args);	

	print "\n</td><td valign='top'>\n\n";
	
	call($args{target},%args);
	
	print "\n</td></tr>\n";
	print "</table>\n";
}
print "<!-- end main -->\n";

htmlfooter();


#######################[ Sub Routines ]############################

sub parse_lastupdate {
	my $file = $_[0];

	my $update = 'n/a';
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		chomp($update = <F>);
		close F;
	}

	return $update;
}

sub parse_usersummary {
	my $file     = $_[0];
	my $username = $_[1] ? $_[1] : '';

	my %info;
	if(-r $file) {
		my $i = 0;

		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			#Thu Jan 13 03:00:01 PST 2005
			#name    round(credit)   total   team
			#anonymous       15743946        437183  0

			if($i >= 2 && $username eq '') {
				my ($name, $credit, $total, $team) = split /\t/;

				$info{$team}{$name}{name}   = $name;
				$info{$team}{$name}{credit} = $credit;
				$info{$team}{$name}{total}  = $total;
				$info{$team}{$name}{team}   = $team;
			}
			elsif($i >= 2 && ($username ne '' && /^\Q$username\E\t/) ) {
				my ($name, $credit, $total, $team) = split /\t/;

				$info{$team}{name}   = $name;
				$info{$team}{credit} = $credit;
				$info{$team}{total}  = $total;
				$info{$team}{team}   = $team;
			}

			$i++;
		}
		close F;
	}

	return %info;
}

sub parse_teamsummary {
	my $file = $_[0];
	my $team = $_[1] ? $_[1] : '';

	my %info;
	if(-r $file) {
		my $i = 0;

		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#Thu Jan 13 03:00:01 PST 2005
			#team    teamname        score   wu
			#0       Default (includes all those WU returned without valid team number)      231453370.603365        6123294
			#446     Google  220443432.95776 4175421

			if($i >= 2 && $team eq '') {
				my ($team, $teamname, $score, $wu) = split /\t/;

				$info{$team}{team}  = $team;
				$info{$team}{name}  = $teamname;
				$info{$team}{score} = $score;
				$info{$team}{wu}    = $wu;
			}
			elsif($i >= 2 && ($team ne '' && /^\Q$team\E\t/) ) {
				my ($team, $teamname, $score, $wu) = split /\t/;
				
				$info{team}  = $team;
				$info{name}  = $teamname;
				$info{score} = $score;
				$info{wu}    = $wu;
			}

			$i++;
		}
		close F;
	}

	return %info;
}

sub parse_projectlist {
	my $file = $_[0];
	
	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
	
			if(/^(.*):(.*?)$/) {
				push @{$info{$1}}, $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_usernamelist {
	my $file = $_[0];

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(/^(.*):(.*?):(.*?)$/) {
				my ($username, $team, $client) = ($1, $2, $3);
			
				push @{$info{$username}{$team}}, $client;
			}
			else {
				print "Error: failed parsing line: $_\n" unless(/^#/);
			}
		}
		close F;
	}

	return %info;
}

sub parse_configinfo {
	my $file = $_[0];

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(/^(.*?)=(.*)$/) {
				$info{$1} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_generalinfo {
	my $file = $_[0];

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			if(/^(.*?):(.*)$/) {
				$info{$1} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_unitinfo {
	my $file = $_[0];
	
	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#Name: p1134_RIBO_FSpeptide_EXT
			if(/^(.*?): (.*)$/) {
				$info{lc($1)} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_psummary {
        my $file = $_[0];
	my $name = $_[1] ? $_[1] : '';
	
	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			# Example line to parse
			#<TR bgcolor="#dfdfdf"> <TD>212</TD><TD> 171.64.122.124 </TD><TD>p212_villinURE212</TD><TD>20605</TD><TD>   27.00</TD>
			#<TD>   39.00</TD><TD>  135.00</TD><TD>100</TD><TD>GROMACS</TD>
			#<TD><a href=http://VSPX27.stanford.edu/cgi-bin/allprojects#212>Description</a></TD>
			#<TD><font size=-1>rbrandman</font></TD></TR>
			
			if(m!<TR bgcolor="#\w{6}">\s*<TD>(\d+)</TD><TD>\s*(\d+\.\d+\.\d+\.\d+)\s*</TD><TD>(.*?)</TD><TD>(\d+)</TD><TD>\s*(\d+\.\d+)</TD><TD>\s*(.*?)</TD><TD>\s*(\d+\.\d+)</TD><TD>(\d+)</TD><TD>(.*?)</TD><TD><a href=(.*?)>Description</a></TD><TD><font size=-1>(.*?)</font></TD></TR>!) {
			
				my ($project_number, $server_ip, $wu_name, $number_of_atoms, $preferred_days, $final_deadline,
				    $credit, $frames, $code, $description_url, $contact) = ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11);
				
				if($name eq '') {
					$info{$wu_name}{project_number}  = $project_number;
					$info{$wu_name}{server_ip}       = $server_ip;
					$info{$wu_name}{wu_name}         = $wu_name;
					$info{$wu_name}{number_of_atoms} = $number_of_atoms;
					$info{$wu_name}{preferred_days}  = $preferred_days;
					$info{$wu_name}{final_deadline}  = $final_deadline;
					$info{$wu_name}{credit}          = $credit;
					$info{$wu_name}{frames}          = $frames;
					$info{$wu_name}{code}            = $code;
					$info{$wu_name}{description_url} = $description_url;
					$info{$wu_name}{contact}         = $contact;
				}
				elsif($name ne '' && $wu_name =~ /^$name/) {
					$info{project_number}  = $project_number;
					$info{server_ip}       = $server_ip;
					$info{wu_name}         = $wu_name;
					$info{number_of_atoms} = $number_of_atoms;
					$info{preferred_days}  = $preferred_days;
					$info{final_deadline}  = $final_deadline;
					$info{credit}          = $credit;
					$info{frames}          = $frames;
					$info{code}            = $code;
					$info{description_url} = $description_url;
					$info{contact}         = $contact;
				}
				
			}
			else {
				#print "ERROR Parsing line: $_";
			}
		}
		close F;
	}
	
	return %info;
}

sub check_directory {
	my ($dir, @perm) = @_;
	my $error = 0;
	my $msg = '';

	foreach(@perm) {
		if($_ eq 'r') {
			if(!-r $dir) {
				$error = 1;
				$msg  .= "Cannot read directory! ($dir)\n";
			}
		}
		elsif($_ eq 'w') {
			if(!-w $dir) {
				$error = 1;
				$msg  .= "Cannot write directory! ($dir)\n";
			}
		}
	}

	return ($error, $msg);
}

sub update_client_list {
	my $file       = $_[0];
	my $client     = $_[1];
	my $new_client = 1;

	if(-r $file && -w $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			if(/^\Q$client\E/) {
				$new_client = 0;
			}
		}
		close F;

		if($new_client == 1) {
			open(F, ">> $file") || die "Cannot open file ($!)";
			print F "$client\n";
			close F;
		}
	}
	else {
		if(!-e $file) {
			# file doesn't exists, create it

			if($new_client == 1) {
				open(F, ">> $file") || die "Cannot open file ($!)";
				print F "$client\n";
				close F;
			}
		}
		else {
			# file does exists, debug
			
			if(!-r $file) {
				print "Error: Cannot read file ($file) !\n";
				return -1;
			}
			elsif(!-w $file) {
				print "Error: Cannot write file ($file) !\n";
				return -1;
			}
			else {
				print "Error: Fatal Exeption on file ($file) !\n";
				return -1;
			}
		}
	}

	return $new_client;
}

sub update_username_list {
	my $file    = $_[0];	#$args{'username-list-file'}
	my $success = 1;
	
	if(-r $file && -w $file) {
		$success = generate_username_list($file);
	}
	else {
		if(!-e $file) {
			# file doesn't exists, create it
			
			$success = generate_username_list($file);
		}
		else {
			# file does exists, debug
			
			if(!-r $file) {
				print "Error: Cannot read file ($file) !\n";
				return -1;
			}
			elsif(!-w $file) {
				print "Error: Cannot write file ($file) !\n";
				return -1;
			}
			else {
				print "Error: Fatal Exeption on file ($file) !\n";
				return -1;
			}
		}
	}

	return $success;
}

sub generate_username_list {
	my $file = $_[0];
	
	my @client;
	open(F, "$args{'client-list-file'}") || die "Cannot open file ($!)";
	while(<F>) {
		chomp;
		
		push @client, $_;
	}
	close F;

	my $data = '';
	foreach(@client) {
		my $client = $_;

		my $config_file = $args{'client-data-dir'}.'client-cfg.'.$client;
		my %configinfo  = parse_configinfo($config_file);

		$data .= "$configinfo{username}:$configinfo{team}:$client\n";
	}

	open(F, "> $file") || die "Cannot open file ($!)";
	print F $data;
	close F;

	return 1;
}

sub update_project_list {
	my $file    = $_[0];
	my $success = 1;
	
	if(-r $file && -w $file) {
		$success = generate_project_list($file);
	}
	else {
		if(!-e $file) {
			# file doesn't exists, create it
			
			$success = generate_project_list($file);
		}
		else {
			# file does exists, debug
			
			if(!-r $file) {
				print "Error: Cannot read file ($file) !\n";
				return -1;
			}
			elsif(!-w $file) {
				print "Error: Cannot write file ($file) !\n";
				return -1;
			}
			else {
				print "Error: Fatal Exeption on file ($file) !\n";
				return -1;
			}
		}
	}

	return $success;
}

sub generate_project_list {
	my $file = $_[0];
	
	my @client;
	open(F, $args{'client-list-file'}) || die "Cannot open file ($!)";
	while(<F>) {
		chomp;
		
		push @client, $_;
	}
	close F;

	my $data = '';
	foreach(@client) {
		my $client = $_;

		my $unitinfo_file = $args{'client-data-dir'}.'unitinfo.'.$client;
		my %unitinfo      = parse_unitinfo($unitinfo_file);
		my %wu_info	  = parse_psummary($args{'project-summary-file'}, $unitinfo{name});

		$data .= "$wu_info{wu_name}:$client\n";
	}

	open(F, "> $file") || die "Cannot open file ($!)";
	print F $data;
	close F;

	return 1;
}

#####[ Start helper subroutines for call() routine ]#####

sub call {
	my $filename = shift;
	my %args = @_;

	my $old_dir = Cwd::cwd;

	my ($file, $dir) = File::Basename::fileparse($filename);

	if($file =~ /\.plc/ && $filename !~ /\//) {
		$dir = Cwd::fast_abs_path($dir);
		
		my $pathname = "$dir/$file";
		Cwd::chdir $dir;

		my $script = parse($pathname);
		if (!$script) {
			print "<div style='color: #FF0000'>Could not load $file</div><br><hr>\n";
			return undef;
		}
	
		my $code = eval $script;
		if ($@ || !$code) {
			print "<div style='color: #FF0000'>Error compiling $file <pre>$@</pre></div><br><hr>\n";
			return undef;
		}
		
		my $result;
		eval {
			$result = &$code(%args);
		};
		
		print "<div style='color: #FF0000'>Error running $file: <pre>$@</pre></div><br><hr>\n" if $@;
		
		Cwd::chdir $old_dir;
		return $result;
	}
	else {
		print "<div style='color: #FF0000'>Cannot load $file</div><br>\n";
		return undef;
	}
}

sub parse {
	my $filename = shift;

	my $script = read_file($filename);
	return undef unless $script;
	
	return 'return sub { my %args = @_; ' . $script . '};';
}

sub read_file {
	my $filename = shift;
	
	my $perl_block = 0;
	my $script;
	
	open(F, $filename) || return undef;
	while (<F>) {
		if (/<\%PERL>/i && !$perl_block) {
			$perl_block = 1;
			$script .= "\n";
		}
		elsif (m|</\%PERL>|i && $perl_block) {
			$perl_block = 0;
			$script .= "\n";
		}
		elsif (/^%(.*)$/ && !$perl_block) {
			$script .= "$1;\n";
		}
		elsif (/<%/ && !$perl_block) {
			while (/<%(.*?)%>/) {
				my $perlblock = $1;
				my $prefix    = $`;
				my $postfix   = $';
				
				$prefix  =~ s/'/'."'".'/g;
				$script .= "print('$prefix');";
				$script .= 'print(' . $perlblock . ');';
				
				$_ = $postfix;
			}
			s/'/'."'".'/g;
			$script .= "print('$_');";
		}
		else {
			if ($perl_block) {
				$script .= $_;
			}
			else {
				s/'/'."'".'/g;
				$script .= "print('$_');";
			}
		}
	}
	close F;
	
	return $script;
}

#####[ End helper subroutines for call() routine ]#####

sub get_args {
	my ($r, $c, $s) = @_;
	my %args;
	
	if ($r->method eq 'POST') {
		my $buffer;
		$r->read($buffer, $r->header_in("Content-length"));
		
		if ($r->header_in("Content-type") =~ m|^multipart/form-data|) {
			my ($boundary) = $r->header_in("Content-type") =~ /boundary=(\S+)/;
			
			my @values = split /--\Q$boundary\E-?-?/, $buffer;	#Escape possible meta-characters with \Q\E
			foreach(@values) {
				next if $_ =~ /^\s*$/;
				
				if(/Content-Disposition: form-data; name="(\S+)"; filename="(.*?)"\r?\n/) {
					my ($name, $filename) = ($1, $2);
					my ($header, $content) = split /\r?\n\r?\n/, $_, 2;
					
					$content =~ s/\r?\n$//;
					
					next if(!$filename || $filename eq '');		# Skip to next when no filename is supplied
					
					my ($mime) = $header =~ /Content-Disposition: form-data; name="$name"; filename=".*?"\s+Content-Type: (\S+)/;

					if($filename =~ /^\w:.*\\(.*?)$/) {
						$filename = $1;
					}

					$args{$name."_Filename"} = $filename;
					$args{$name."_Mime"}	 = $mime;
					$args{$name}		 = $content;
				}
				else {
					# Regular URL args key=value
					
					my ($header, $content) = split /\r?\n\r?\n/, $_, 2;
					$content =~ s/\r?\n$//;
					my ($name) = $header =~ /Content-Disposition: form-data; name="(\S+)"/;

					$args{$name} = $content;
				}
			}
		}
		else {
			my $i = 0;
			foreach (split ('&', $buffer)) {
				my ($param, $value) = split '=';
				$param =~ tr/+/ /;
				$param =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
				$value =~ tr/+/ /;
				$value =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
			
				$args{$param} = '' if(!$args{$param}); #Fix for 'Use of uninitialized value in string eq at [..]' error message
			
				if($args{$param} eq ''){
					$args{$param} = $value;
					$i=0;
				}
				else{
					$args{$param."_".$i}=$value;
					$i++;
				}
			}
		}
	}
	else {
		my $i = 0;
		foreach (split ('&', $r->query_string)) {
			my ($param, $value) = split '=';
			$param =~ tr/+/ /;
			$param =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
			$value =~ tr/+/ /;
			$value =~ s/%([0-9a-fA-F]{2})/pack("c",hex($1))/ge;
			
			$args{$param} = '' if(!$args{$param});  # Small hack to avoid 'Use of uninitialized value in string eq at [..]' error message
			if($args{$param} eq ''){
				$args{$param} = $value;
				$i=0;
			}
			else{
				$args{$param."_".$i}=$value;
				$i++;
			}
		}
	}

	my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();
	
	$month = '0'.$month if(length($month) == 1);
	$day   = '0'.$day   if(length($day) == 1);
	$hour  = '0'.$hour  if(length($hour) == 1);
	$min   = '0'.$min   if(length($min) == 1);
	$sec   = '0'.$sec   if(length($sec) == 1);

	$args{"User"}           = $c->user                          if defined($c->user);
	$args{"Date"}		= "$year-$month-$day";
	$args{"Time"}		= "$hour:$min:$sec";
	$args{"Cookie"}     	= $r->header_in('Cookie') 	    if($r->header_in('Cookie'));
	$args{"Referer"}    	= $r->header_in('Referer') 	    if($r->header_in('Referer'));
	$args{"Protocol"}   	= $r->protocol;
	$args{"UserAgent"}  	= $r->header_in('User-Agent');
	$args{"Host"}       	= $r->header_in('Host');
	$args{"Accept"}     	= $r->header_in('Accept');
	$args{"AcceptLanguage"} = $r->header_in('Accept-Language');
	$args{"AcceptEncoding"} = $r->header_in('Accept-Encoding');
	$args{"AcceptCharset"}  = $r->header_in('Accept-Charset');
	$args{"Connection"}     = $r->header_in('Connection');
	$args{"Authorization"}  = $r->header_in('Authorization')    if($r->header_in('Authorization'));
	$args{"Pragma"}         = $r->header_in('Pragma') 	    if($r->header_in('Pragma'));
	$args{"CacheControl"}   = $r->header_in('Cache-Control')    if($r->header_in('Cache-Control'));

	$args{"Hostname"}   	= $r->hostname;
	$args{"RemoteIP"}	= $c->remote_ip();
	if($c->remote_host() eq '') {
		$args{"RemoteHostname"} = $r->get_remote_host();
	}
	else {
		$args{"RemoteHostname"} = $c->remote_host();
	}

	$args{"ServerName"}     = $s->server_hostname;
	$args{"ServerAdmin"}    = $s->server_admin;
	$args{"ServerPort"}     = $s->port;
	$args{"ServerSoftware"} = $ENV{SERVER_SOFTWARE};	

	return %args;
}

sub htmlheader {
	my %args = @_;
	
	if($args{doctype}) {
		print "<!DOCTYPE $args{doctype} PUBLIC \"-//W3C//DTD ";
		if($args{doctype_dtd}) {
			print "$args{doctype_dtd} ";
		}
		else {
			print "$args{doctype} ";
		}
		
		if($args{doctype_transitional} == 1) {
			print "Transitional";
		}
		elsif($args{doctype_strict} == 1) {
			print "Strict";
		}
		elsif($args{doctype_loose} == 1) {
			print "Loose";
		}
		else {
			print "Transitional";
		}
		
		if($args{doctype_language}) {
			print "//$args{doctype_language}\">\n";
		}
		else {
			print "//EN\">\n";
		}
	}
	
	print "<html>\n";
	print "<head>\n";
	print "\t<title>$args{title}</title>\n" if($args{title});
	if($args{meta1}) {
		my $i = 1;
		while($args{'meta' . $i}) {
			print "\t<meta";
			
			my $meta_name = $args{'meta' . $i . '_name'} if($args{'meta' . $i . '_name'});
			print " name=\"$meta_name\"" if($args{'meta' . $i . '_name'});
			
			my $meta_equiv = $args{'meta' . $i . '_equiv'} if($args{'meta' . $i . '_equiv'});
			print " http-equiv=\"$meta_equiv\"" if($args{'meta' . $i . '_httpequiv'});
			
			my $content  = $args{'meta' . $i . '_content'} if($args{'meta' . $i . '_content'});
			print " content=\"$content\"" if($args{'meta' . $i . '_content'});
			
			$i++;
		}
	}
	if($args{link1}) {
		my $i = 1;
		while($args{'link' . $i}) {
			print "\t<link";
			
			my $linktype = $args{'link' . $i . '_type'} if($args{'link' . $i . '_type'});
			print " type=\"$linktype\"" if($args{'link' . $i . '_type'});
			
			my $linkrel  = $args{'link' . $i . '_rel'} if($args{'link' . $i . '_rel'});
			print " rel=\"$linkrel\"" if($args{'link' . $i . '_rel'});
			
			print " href=\"$args{'link' . $i}\">\n";
			
			$i++;
		}
	}
	if($args{script1}) {
		my $i = 1;
		while($args{'script' . $i}) {
			print "\t<script";
			
			my $scripttype = $args{'script' . $i . '_type'} if($args{'script' . $i . '_type'});
			print " type=\"$scripttype\"" if($args{'script' . $i . '_type'});
			
			my $scriptlanguage  = $args{'script' . $i . '_language'} if($args{'script' . $i . '_language'});
			print " language=\"$scriptlanguage\"" if($args{'script' . $i . '_language'});
			
			print " src=\"$args{'script' . $i}\"></script>\n";
			
			$i++;
		}
	}
	print "\t<style>$args{style}</style>\n" if($args{style});
	
	print "</head>\n\n";
	
	print "<body";
	print " onUnload=\"$args{body_onunload}\"" if($args{body_onunload});
	print " onLoad=\"$args{body_onload}\"" if($args{body_onload});
	print ">\n\n";
}

sub htmlfooter {
	print "\n</body>\n";
	print "</html>\n";
}
